@@ -41,7 +41,6 @@
 #undef HAVE_FILE_FUNCS
 #endif
 
-#include "mp4h.h"
 #include "regex.h"
 
 #ifdef HAVE_FILE_FUNCS
@@ -57,8 +56,11 @@
 #include <locale.h>
 #include <time.h>
 
-#define _AS_HEADER
+#include "mp4h.h"
+
+#define _SRC_VERSION_C_AS_HEADER_
 #include "version.c"
+#undef _SRC_VERSION_C_AS_HEADER_
 
 #define MP4H_BUILTIN_ARGS struct obstack *obs, int argc, token_data **argv, \
                             read_type expansion
