@@ -133,8 +133,8 @@ struct input_block
       u_c;
       struct
         {
-          char *start; /* string value */
-          char *current; /* current value */
+          unsigned char *start;   /* string value */
+          unsigned char *current; /* current value */
         }
       u_s;
       struct
@@ -143,8 +143,8 @@ struct input_block
           char *name;           /* name of PREVIOUS input file */
           int lineno;           /* current line number for do */
           /* Yet another attack of "The curse of global variables" (sigh) */
-          int out_lineno;        /* current output line number do */
-          boolean advance_line;  /* start_of_input_line from next_char () */
+          int out_lineno;       /* current output line number do */
+          boolean advance_line; /* start_of_input_line from next_char () */
         }
       u_f;
       struct
@@ -274,7 +274,9 @@ file_clean(void)
                     isp->u.u_f.name, isp->u.u_f.lineno);
 
   fclose (isp->u.u_f.file);
-  current_file = isp->u.u_f.name;
+  xfree (current_file);
+  current_file = xstrdup (isp->u.u_f.name);
+  xfree (isp->u.u_f.name);
   current_line = isp->u.u_f.lineno;
   output_current_line = isp->u.u_f.out_lineno;
   start_of_input_line = isp->u.u_f.advance_line;
@@ -310,7 +312,7 @@ push_file (FILE *fp, const char *title)
   i->u.u_f.out_lineno = output_current_line;
   i->u.u_f.advance_line = start_of_input_line;
 
-  current_file = obstack_copy0 (current_input, title, strlen (title));
+  current_file = xstrdup (obstack_copy0 (current_input, title, strlen (title)));
   current_line = 1;
   output_current_line = -1;
 
@@ -494,15 +496,15 @@ push_string_finish (read_type expansion)
           || expansion == READ_BODY)
         {
           TOKEN_DATA_TYPE (&token_read) = TOKEN_TEXT;
-          TOKEN_DATA_TEXT (&token_read) = xstrdup (next->u.u_s.start);
-          next->u.u_s.current = next->u.u_s.start + strlen (next->u.u_s.start);
+          TOKEN_DATA_TEXT (&token_read) = xstrdup ((char *) next->u.u_s.start);
+          next->u.u_s.current = next->u.u_s.start + strlen ((char *) next->u.u_s.start);
         }
       else
         next->u.u_s.current = next->u.u_s.start;
 
       next->prev = isp;
       isp = next;
-      ret = isp->u.u_s.start;   /* for immediate use only */
+      ret = (char *) isp->u.u_s.start;   /* for immediate use only */
     }
   else
     obstack_free (current_input, next); /* people might leave garbage on it. */
@@ -736,6 +738,9 @@ skip_line (void)
 void
 skip_buffer (void)
 {
+  if (isp == NULL)
+    return;
+
   pop_input ();
 }
 
@@ -752,11 +757,11 @@ skip_buffer (void)
 `---------------------------------------------------------------------*/
 
 static boolean
-match_comment (const char *s)
+match_comment (const unsigned char *s)
 {
   int n;                        /* number of characters matched */
   int ch;                       /* input character */
-  const char *t;
+  const unsigned char *t;
   struct obstack *st;
 
   ch = peek_input ();
@@ -777,7 +782,7 @@ match_comment (const char *s)
   /* Push back input.  */
   st = push_string_init ();
   obstack_grow (st, t, n);
-  push_string_finish (0);
+  push_string_finish (READ_NORMAL);
   return FALSE;
 }
 
@@ -794,7 +799,7 @@ match_comment (const char *s)
   ((s)[0] == (ch) \
    && (ch) != '\0' \
    && ((s)[1] == '\0' \
-       || (match_comment ((s) + 1))))
+       || (match_comment ((unsigned char *) (s) + 1))))
 
 
 
@@ -805,9 +810,7 @@ match_comment (const char *s)
 void
 input_init (void)
 {
-  int ch;
-
-  current_file = _("NONE");
+  current_file = xstrdup ("NONE");
   current_line = 0;
   array_current_line = (int *) xmalloc (sizeof (int) * nesting_limit);
 
@@ -836,6 +839,21 @@ input_init (void)
   rquote.length  = strlen (rquote.string);
   visible_quotes = FALSE;
 
+  /* this function is defined in builtin.c and initializes the
+     varbreak symbol.  */
+  init_break ();
+
+  /* This routine tells whether we want extended regexp or not. */
+  set_regexp_extended (TRUE);
+  
+  syntax_init ();
+}
+
+void
+syntax_init (void)
+{
+  int ch;
+
   for (ch = 256; --ch > 0; )
     {
       if (isspace(ch))
@@ -853,14 +871,10 @@ input_init (void)
   set_syntax_internal(SYNTAX_CLOSE,  '>');
   set_syntax_internal(SYNTAX_ACTIVE, '\\');
   set_syntax_internal(SYNTAX_ACTIVE, '"');
-
-  /* this function is defined in builtin.c and initializes the
-     varbreak symbol.  */
-  init_break ();
-
-  /* This routine tells whether we want extended regexp or not. */
-  set_regexp_extended (TRUE);
-  
+  set_syntax_internal(SYNTAX_GROUP, CHAR_LQUOTE);
+  set_syntax_internal(SYNTAX_GROUP, CHAR_RQUOTE);
+  set_syntax_internal(SYNTAX_GROUP, CHAR_BGROUP);
+  set_syntax_internal(SYNTAX_GROUP, CHAR_EGROUP);
 }
 
 /*---------------------------------------------------------.
@@ -1040,7 +1054,7 @@ next_token (token_data *td, read_type expansion, boolean in_string)
                       else
                         unget_input(ch);
 
-                      if (IS_SPACE(ch) || IS_CLOSE(ch) || ch == '/')
+                      if (IS_SPACE(ch) || IS_CLOSE(ch) || IS_SLASH (ch))
                         type = TOKEN_WORD;
                       else
                         type = TOKEN_STRING;
@@ -1062,7 +1076,8 @@ next_token (token_data *td, read_type expansion, boolean in_string)
       else if (expansion == READ_BODY)
         {
           obstack_1grow (&token_stack, ch);
-          while ((ch = next_char ()) != CHAR_EOF && ! IS_ESCAPE(ch))
+          while ((ch = next_char ()) != CHAR_EOF
+                  && ! IS_ESCAPE(ch) && ! IS_CLOSE (ch))
             {
               if (eolcomm.length > 0 && MATCH (ch, eolcomm.string))
                 {
@@ -1070,7 +1085,7 @@ next_token (token_data *td, read_type expansion, boolean in_string)
                   ch = CHAR_EOF;
                   break;
                 }
-                obstack_1grow (&token_stack, ch);
+              obstack_1grow (&token_stack, ch);
             }
           unget_input(ch);
 
@@ -1219,7 +1234,12 @@ next_token (token_data *td, read_type expansion, boolean in_string)
           if (IS_OTHER(ch) || IS_NUM(ch))
             type = TOKEN_STRING;
           else if (IS_SPACE(ch))
-            type = TOKEN_SPACE;
+            {
+              while ((ch = next_char ()) != CHAR_EOF && IS_SPACE(ch))
+                obstack_1grow (&token_stack, ch);
+              unget_input(ch);
+              type = TOKEN_SPACE;
+            }
           else if (IS_ACTIVE(ch))
             type = TOKEN_WORD;
           else
@@ -1249,10 +1269,27 @@ void
 read_file_verbatim (struct obstack *obs)
 {
   int ch;
+  int (*f)(void);
+  struct obstack *st;
 
-  while ((ch = file_read ()) != CHAR_RETRY)
-    shipout_text (0, (char *) &ch, 1);
-  pop_input ();
+  if (next != NULL)
+    st = obs;
+  else
+    st = push_string_init ();
+
+  f = isp->funcs->read_func;
+  if (f != NULL)
+    {
+      while ((ch = (*f)()) != CHAR_RETRY)
+        obstack_1grow (st, ch);
+    }
+  else
+    {
+      MP4HERROR ((warning_status, 0,
+        _("INTERNAL ERROR: Input stack botch in read_file_verbatim ()")));
+      abort ();
+    }
+  push_string_finish (READ_BODY);
 }
 
 
