@@ -89,7 +89,7 @@ expand_token (struct obstack *obs, read_type expansion, token_type t,
     case TOKEN_STRING:
       if (expansion_level > 0 && t == TOKEN_QUOTED)
         obstack_1grow (obs, CHAR_LQUOTE);
-      shipout_text (obs, text, strlen (text));
+      shipout_text (obs, text);
       if (expansion_level > 0 && t == TOKEN_QUOTED)
         obstack_1grow (obs, CHAR_RQUOTE);
       break;
@@ -108,8 +108,7 @@ expand_token (struct obstack *obs, read_type expansion, token_type t,
           if (sym == NULL || SYMBOL_TYPE (sym) == TOKEN_VOID)
             {
               if (exp_flags & EXP_NO_HTMLTAG)
-                shipout_text (obs, TOKEN_DATA_TEXT (td),
-                        strlen (TOKEN_DATA_TEXT (td)));
+                shipout_text (obs, TOKEN_DATA_TEXT (td));
               else
                 expand_unknown_tag (text, expansion);
             }
@@ -117,7 +116,7 @@ expand_token (struct obstack *obs, read_type expansion, token_type t,
             expand_macro (sym, expansion);
         }
       else
-        shipout_text (obs, TOKEN_DATA_TEXT (td), strlen (TOKEN_DATA_TEXT (td)));
+        shipout_text (obs, TOKEN_DATA_TEXT (td));
       break;
 
     default:
@@ -152,11 +151,18 @@ expand_argument (struct obstack *obs, read_type expansion, token_data *argp,
   TOKEN_DATA_TYPE (argp) = TOKEN_VOID;
 
   /* Skip leading white space.  */
-  do
+  t = next_token (&td, expansion, FALSE);
+  if (t == TOKEN_SPACE)
     {
+      if (expansion == READ_ATTR_ASIS || expansion == READ_ATTR_QUOT)
+        {
+          /*  We are parsing attributes of a command which will be
+              rescanned, so whitepsaces are preserved.  */
+          obstack_grow (obs, TOKEN_DATA_TEXT (&td),
+                  strlen (TOKEN_DATA_TEXT (&td)));
+        }
       t = next_token (&td, expansion, FALSE);
     }
-  while (t == TOKEN_SPACE);
 
   while (1)
     {
@@ -170,6 +176,13 @@ expand_argument (struct obstack *obs, read_type expansion, token_data *argp,
                && !in_string && (group_level == 0))
             {
 
+              if (t == TOKEN_SPACE &&
+                  (expansion == READ_ATTR_ASIS || expansion == READ_ATTR_QUOT))
+                {
+                  obstack_grow (obs, TOKEN_DATA_TEXT (&td),
+                          strlen (TOKEN_DATA_TEXT (&td)));
+                }
+
               /* The argument MUST be finished, whether we want it or not.  */
               obstack_1grow (obs, '\0');
 
@@ -267,7 +280,7 @@ collect_arguments (char *symbol_name, read_type expansion,
   ch = peek_input ();
   if (IS_CLOSE (ch))
     (void) next_token (&td, READ_BODY, FALSE);
-  else if (IS_SPACE(ch) || IS_ESCAPE(ch) || ch == '/')
+  else if (IS_SPACE(ch) || IS_ESCAPE(ch) || IS_SLASH(ch))
     {
       do
         {
@@ -281,12 +294,13 @@ collect_arguments (char *symbol_name, read_type expansion,
                 _("ERROR:%s:%d: EOF when reading argument of the `%s' tag"),
                      CURRENT_FILE_LINE, symbol_name));
             }
+
           tdp = (token_data *)
             obstack_copy (arguments, (voidstar) &td, sizeof (td));
           obstack_grow (argptr, (voidstar) &tdp, sizeof (tdp));
         }
       while (more_args);
-      
+
       /*  If the last argument is empty, it is removed.  We need it to
           remove white spaces before closing brackets.  */
       if (TOKEN_DATA_TYPE (tdp) == TOKEN_TEXT
@@ -299,7 +313,7 @@ collect_arguments (char *symbol_name, read_type expansion,
         _("INTERNAL ERROR: Bad tag expression in `%s'"),
              CURRENT_FILE_LINE, symbol_name));
     }
-  return (last_char == '/');
+  return (IS_SLASH(last_char));
 }
 
 /*-----------------------------------------------------------------.
@@ -344,7 +358,7 @@ collect_body (char *symbol_name, read_type expansion,
         case TOKEN_STRING:
           if (expansion_level > 0 && t == TOKEN_QUOTED)
             obstack_1grow (bodyptr, CHAR_LQUOTE);
-          shipout_text (bodyptr, text, strlen (text));
+          shipout_text (bodyptr, text);
           if (expansion_level > 0 && t == TOKEN_QUOTED)
             obstack_1grow (bodyptr, CHAR_RQUOTE);
           break;
@@ -421,8 +435,7 @@ collect_body (char *symbol_name, read_type expansion,
               if (newsym == NULL || SYMBOL_TYPE (newsym) == TOKEN_VOID)
                 {
                   if (exp_flags & EXP_NO_HTMLTAG)
-                    shipout_text (bodyptr, TOKEN_DATA_TEXT (&td),
-                            strlen (TOKEN_DATA_TEXT (&td)));
+                    shipout_text (bodyptr, TOKEN_DATA_TEXT (&td));
                   else
                     expand_unknown_tag (text, expansion);
                 }
@@ -541,10 +554,15 @@ expand_macro (symbol *sym, read_type expansion)
       if (slash)
         {
           cp = TOKEN_DATA_TEXT (argv[argc-1]);
-          if (*cp == '/' && *(cp+1) == '\0')
+          if (IS_SLASH(*cp) && *(cp+1) == '\0')
             argc--;
-          else if (LAST_CHAR (cp) == '/')
-            LAST_CHAR (cp) = '\0';
+          else if (IS_SLASH (LAST_CHAR (cp)))
+            {
+              if (IS_SPACE (*(cp+strlen(cp)-2)))
+                *(cp+strlen(cp)-2) = '\0';
+              else
+                LAST_CHAR (cp) = '\0';
+            }
         }
       else if (!(exp_flags & EXP_NOWARN_SLASH))
         MP4HERROR ((warning_status, 0,
@@ -572,39 +590,11 @@ expand_macro (symbol *sym, read_type expansion)
 
       for (i = 1; i < argc; i++)
         {
-          obstack_1grow (obs_expansion, ' ');
-          if (expansion == READ_ATTR_ASIS)
-            shipout_string (obs_expansion, TOKEN_DATA_TEXT (argv[i]), 0);
-          else
-            {
-              /*  Attributes are put back on stack.
-                  The first idea is to write
-                    obstack_1grow (obs_expansion, CHAR_BGROUP);
-                    shipout_string (obs_expansion, TOKEN_DATA_TEXT (argv[i]), 0);
-                    obstack_1grow (obs_expansion, CHAR_EGROUP);
-                  But this does not work because if string is `%attributes'
-                  then further expansion will be wrong.
-              */
-              if (strlen (TOKEN_DATA_TEXT (argv[i])) == 0)
-                {
-                  obstack_1grow (obs_expansion, CHAR_BGROUP);
-                  obstack_1grow (obs_expansion, CHAR_EGROUP);
-                }
-              else if (strchr (TOKEN_DATA_TEXT (argv[i]), ' '))
-                {
-                  obstack_1grow (obs_expansion, CHAR_BGROUP);
-                  shipout_string (obs_expansion, TOKEN_DATA_TEXT (argv[i]), 0);
-                  obstack_1grow (obs_expansion, CHAR_EGROUP);
-                }
-              else
-                shipout_string (obs_expansion, TOKEN_DATA_TEXT (argv[i]), 0);
-            }
+          shipout_string (obs_expansion, TOKEN_DATA_TEXT (argv[i]), 0);
         }
-      if (!SYMBOL_CONTAINER (sym))
+      if (!SYMBOL_CONTAINER (sym) && slash)
         {
-          if (argc > 1)
-            obstack_1grow (obs_expansion, ' ');
-          obstack_1grow (obs_expansion, '/');
+          obstack_1grow (obs_expansion, CHAR_SLASH);
         }
 
       obstack_1grow (obs_expansion, '>');
@@ -646,7 +636,7 @@ expand_unknown_tag (char *name, read_type expansion)
   const char *expanded;
   char *symbol_name, *cp;
   read_type attr_expansion;
-  boolean slash;
+  boolean slash, single;
 
   expansion_level++;
 
@@ -656,7 +646,8 @@ expand_unknown_tag (char *name, read_type expansion)
   obstack_init (&argptr);
   obstack_init (&body);
 
-  if (expansion == READ_NORMAL || expansion == READ_ATTRIBUTE)
+  if (expansion == READ_NORMAL || expansion == READ_ATTRIBUTE
+          || expansion == READ_ATTR_QUOT)
     attr_expansion = READ_ATTR_QUOT;
   else
     attr_expansion = READ_ATTR_ASIS;
@@ -670,26 +661,31 @@ expand_unknown_tag (char *name, read_type expansion)
          - tag name last character is a star
          - attributes are ended by a trailing slash
   */
+  single = slash;
   if (*(symbol_name) == '/')
-    slash = TRUE;
+    single = TRUE;
 
-  if (!slash && !(exp_flags & EXP_STAR_COMPLEX))
-    slash = (LAST_CHAR (symbol_name) == '*');
+  if (!single && !(exp_flags & EXP_STAR_COMPLEX))
+    single = (LAST_CHAR (symbol_name) == '*');
 
-  if (!slash && !(exp_flags & EXP_DFT_SIMPLE))
+  if (!single && !(exp_flags & EXP_DFT_SIMPLE))
     collect_body (symbol_name, expansion, &argptr, &body);
 
   argv = (token_data **) obstack_finish (&argptr);
 
-  /*  When this tag is no more processed, remove the trailing slash.  */
-  if ((exp_flags & EXP_REMOVE_TRAILING_SLASH) &&
-      expansion == READ_NORMAL && slash)
+  /*  Remove the trailing slash in single tags.  */
+  if (slash)
     {
       cp = TOKEN_DATA_TEXT (argv[argc-1]);
-      if (*cp == '/' && *(cp+1) == '\0')
+      if (IS_SLASH (*cp) && *(cp+1) == '\0')
         argc--;
-      else if (LAST_CHAR (cp) == '/')
-        LAST_CHAR (cp) = '\0';
+      else if (IS_SLASH (LAST_CHAR (cp)))
+        {
+          if (IS_SPACE (*(cp+strlen(cp)-2)))
+            *(cp+strlen(cp)-2) = '\0';
+          else
+            LAST_CHAR (cp) = '\0';
+        }
     }
 
   if (expansion == READ_ATTR_ASIS
@@ -703,15 +699,20 @@ expand_unknown_tag (char *name, read_type expansion)
   obstack_1grow (obs_expansion, '<');
   shipout_string (obs_expansion, symbol_name, 0);
 
+  /*  Whitespaces between attributes have been put into
+      attributes text, so there is no need to insert a space
+      between them.  */
   for (i = 1; i < argc; i++)
+    shipout_string (obs_expansion, TOKEN_DATA_TEXT (argv[i]), 0);
+
+  if (slash)
     {
-      obstack_1grow (obs_expansion, ' ');
-      shipout_string (obs_expansion, TOKEN_DATA_TEXT (argv[i]), 0);
+      obstack_1grow (obs_expansion, CHAR_SLASH);
     }
   obstack_1grow (obs_expansion, '>');
   if (expansion != READ_BODY)
     obstack_1grow (obs_expansion, CHAR_RQUOTE);
-  if (!slash && !(exp_flags & EXP_DFT_SIMPLE))
+  if (!single && !(exp_flags & EXP_DFT_SIMPLE))
     {
       shipout_string (obs_expansion, TOKEN_DATA_TEXT (argv[argc]), 0);
       if (expansion != READ_BODY)
