@@ -33,8 +33,9 @@
 
 #include <getopt.h>
 
-#define _AS_HEADER
+#define _SRC_VERSION_C_AS_HEADER_
 #include "version.c"
+#undef _SRC_VERSION_C_AS_HEADER_
 
 /* Enable sync output for /lib/cpp (-s).  */
 int sync_output = 0;
@@ -89,22 +90,6 @@ struct macro_definition
 typedef struct macro_definition macro_definition;
 
 
-
-/* Memory allocation.  */
-
-/*------------------------.
-| Failsafe free routine.  |
-`------------------------*/
-
-void
-xfree (voidstar p)
-{
-  if (p != NULL)
-    free (p);
-  p = NULL;
-}
-
-
 /*---------------------------------------------.
 | Print a usage message and exit with STATUS.  |
 `---------------------------------------------*/
@@ -362,7 +347,7 @@ main (int argc, char *const *argv)
 
   if (show_version)
     {
-      printf ("%s %s", program_name, mp4h_VersionStr);
+      printf ("%s %s", program_name, mp4h_version.v_long);
       fputs("\n", stdout);
       exit (EXIT_SUCCESS);
     }
@@ -374,6 +359,7 @@ main (int argc, char *const *argv)
 
   input_init ();
   output_init ();
+  include_env_init ();
   symtab_init ();
   locale_init ();
 
@@ -450,6 +436,7 @@ main (int argc, char *const *argv)
               }
             else
               {
+                xfree (current_file);
                 current_file = xstrdup (filename);
                 push_file (fp, filename);
                 xfree (filename);
