@@ -369,6 +369,7 @@ typedef int read_type;
 
 void input_init __P ((void));
 void input_deallocate __P ((void));
+void syntax_init __P ((void));
 int peek_input __P ((void));
 token_type next_token __P ((token_data *, read_type, boolean));
 void skip_line __P ((void));
@@ -400,11 +401,15 @@ extern STRING lquote, rquote;
 /* Eof-of-line comment  */
 extern STRING eolcomm;
 
+/* Special characters used for grouping  */
 #define CHAR_LQUOTE '\1'
 #define CHAR_RQUOTE '\2'
 #define CHAR_BGROUP '\3'
 #define CHAR_EGROUP '\4'
+
+/* Some characters are replaced during input/output phases */
 #define CHAR_QUOTE  '\5'
+#define CHAR_SLASH  '\6'
 
 /* Default eof-of-line comment  */
 #define DEF_EOLCOMM  ";;;"
@@ -425,6 +430,7 @@ extern unsigned short syntax_table[256];
 #define SYNTAX_CLOSE    (0x0004)
 #define SYNTAX_ACTIVE   (0x0007)
 #define SYNTAX_ESCAPE   (0x0008)
+#define SYNTAX_GROUP    (0x0009)
 
 /* These are values to be assigned to syntax table entries, but they are
    used as bit masks with IS_ALNUM.*/
@@ -451,6 +457,8 @@ extern unsigned short syntax_table[256];
 #define IS_EGROUP(ch) (ch == CHAR_EGROUP)
 #define IS_LQUOTE(ch) (ch == CHAR_LQUOTE)
 #define IS_RQUOTE(ch) (ch == CHAR_RQUOTE)
+#define IS_GROUP(ch)  ((syntax_table[(int)(ch)]&SYNTAX_VALUE) == SYNTAX_GROUP)
+#define IS_SLASH(ch)  (ch == CHAR_SLASH || ch == '/')
 
 void set_syntax __P ((int, const char *));
 void set_syntax_internal __P ((int, int));
@@ -463,13 +471,13 @@ extern int output_current_line;
 
 void output_init __P ((void));
 void output_deallocate __P ((void));
-void shipout_text __P ((struct obstack *, char *, int));
+void shipout_text __P ((struct obstack *, char *));
 void shipout_string __P ((struct obstack *obs, const char *s, int len));
 void make_diversion __P ((int));
 void insert_diversion __P ((int));
 void insert_file __P ((FILE *));
 void freeze_diversions __P ((FILE *));
-int remove_special_chars __P ((char *, boolean));
+void remove_special_chars __P ((char *, boolean));
 
 
 /* File symtab.c  --- symbol table definitions.  */
@@ -513,12 +521,16 @@ typedef void hack_symbol ();
 
 #define HASHMAX 509             /* default, overridden by -Hsize */
 
+extern symbol **sym_tab;
+extern symbol **var_tab;
+extern symbol **file_tab;
 extern symbol **symtab;
 
 void symtab_init __P ((void));
 void symtab_deallocate __P ((void));
 symbol *lookup_symbol __P ((const char *, symbol_lookup));
 symbol *lookup_variable __P ((const char *, symbol_lookup));
+symbol *lookup_file __P ((const char *, symbol_lookup));
 void hack_all_symbols __P ((hack_symbol *, const char *));
 
 
