@@ -58,18 +58,45 @@ search_path_add (struct search_path_info *info, const char *dir)
   info->list_end = path;
 }
 
+static void
+search_path_env_init (struct search_path_info *info, const char *path)
+{
+  char *path_end;
+
+  if (info == NULL || path == NULL)
+    return;
+
+  do
+    {
+      path_end = strchr (path, ':');
+      if (path_end)
+        *path_end = '\0';
+      search_path_add (info, path);
+      path = path_end + 1;
+    }
+  while (path_end);
+}
+
 
 void
 include_init (void)
 {
   dirpath.list = NULL;
   dirpath.list_end = NULL;
-  dirpath.max_length = 0;
+  /* This directory is always searched for packages */
+  dirpath.max_length = strlen (PKGDATADIR);
 }
 
 
 /* Functions for normal input path search */
 
+void
+include_env_init (void)
+{
+  search_path_env_init (&dirpath, getenv ("MP4HPATH"));
+}
+
+
 void
 add_include_directory (const char *dir)
 {
@@ -101,7 +128,6 @@ path_search (const char *dir, char **expanded_name)
     return NULL;
 
   name = (char *) xmalloc (dirpath.max_length + 1 + strlen (dir) + 1);
-
   for (incl = dirpath.list; incl != NULL; incl = incl->next)
     {
       strncpy (name, incl->dir, incl->len);
@@ -124,6 +150,27 @@ path_search (const char *dir, char **expanded_name)
         }
     }
 
+  if (fp == NULL)
+    {
+      strcpy (name, PKGDATADIR);
+      name[strlen (PKGDATADIR)] = '/';
+      strcpy (name + strlen (PKGDATADIR) + 1, dir);
+
+#ifdef DEBUG_INCL
+      fprintf (stderr, "path_search (%s) -- trying %s\n", dir, name);
+#endif
+
+      fp = fopen (name, "r");
+      if (fp != NULL)
+        {
+          if (debug_level & DEBUG_TRACE_PATH)
+            DEBUG_MESSAGE2 (_("Path search for `%s' found `%s'"), dir, name);
+
+          if (expanded_name != NULL)
+            *expanded_name = xstrdup (name);
+        }
+    }
+
   xfree (name);
 
   return fp;
